/**
 ** File:   libtcpip.c
 ** Date:   Sun Oct 15 1995
 ** Author: Merijn de Jonge
 ** Email:  mdejonge@fwi.uva.nl
 **
 ** This file is part of the modem communication package.
 **
 ** Copyright (C) 1995 Merijn de Jonge.
 **
 ** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 ** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 ** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 ** 
 ** Redistribution and use in source and binary forms are permitted provided
 ** that: (1) source distributions retain this entire copyright notice and
 ** comment, and (2) distributions may not be sold for profit on physical
 ** media such as disks, tapes, and CD-ROMS, without expressed written
 ** permission.
 ** 
 **/
#include <stdlib.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "libtcpip.h"

#define UDP_PROT "udp"
#define TCP_PROT "tcp"
#define BACKLOG	3

int sockSetProtocol( sockStruct* sock, sockProtocol prot )
{
	if( sock == NULL )
		return -1;
	sock->prot = prot;
	return 0;
}

int sockSetPort( sockStruct* sock, int port )
{
	if( sock == NULL )
		return -1;
	sock->port = htons( port );
	return 0;
}

int sockSetPortToService( sockStruct* sock, char* service )
{
	struct servent *serv;
	char* prot;
	
	switch( sock->prot )
	{
		case UDP:
			prot = UDP_PROT;
			break;
		case TCP:
			prot = TCP_PROT;
			break;
		default:
			return -1;
	}	
	if( sock == 0 )
		return -1;
	
	serv = getservbyname( service, prot );
	if( serv == NULL )
		return -1;
	sock->port = serv->s_port;
	return 0;
}

static int sockMakeSocket( sockStruct* sock )
{
	int type;
	if( sock == NULL )
		return -1;
	switch( sock->prot )
	{
		case UDP:
			type = SOCK_DGRAM;
			break;
		case TCP:
			type = SOCK_STREAM;
			break;
		default:
			return -1;
	}
	sock->fd = socket( AF_INET, type, 0 );
	if( sock->fd < 0 )
		return -1;
	sock->name.sin_family = AF_INET;
	sock->name.sin_port = sock->port;

	return 0;
}

int sockServer( sockStruct* sock )
{
	if( sockMakeSocket( sock ) == -1 )
		return -1;
	sock->name.sin_addr.s_addr = htonl( INADDR_ANY );
	if( bind( sock->fd, (struct sockaddr *)&(sock->name), sizeof(
		sock->name) ) < 0 )
		return -1;
	return 0;
}

int sockClient( sockStruct* sock, const char* host )
{
	struct hostent *hostinfo;
	
	if( sockMakeSocket( sock ) == -1 )
		return -1;
	
	hostinfo = gethostbyname( host );
	if( host == NULL )
		return -1;
	sock->name.sin_addr = *(struct in_addr *)hostinfo->h_addr;
	return 0;
}

int  sockSetOption( sockStruct* sock, int option, int _on )
{
	int on = _on;
	if( sock == NULL )
		return -1;
	return setsockopt( sock->fd, SOL_SOCKET, option, &on, sizeof( on ) );
}

int sockListen( sockStruct* sock )
{
	if( sock == NULL )
		return -1;
	return listen( sock->fd, BACKLOG );
}

int sockAccept( sockStruct* sock, sockStruct* client )		
{
	int size;
	
	if( sock == NULL || client == NULL )
		return -1;
		
	size = sizeof( client->name);
	client->fd = accept( sock->fd, (struct sockaddr*)&(client->name), &size );
	
	if( client->fd == -1 )
		return -1;
	return 0;
}

int sockConnect( sockStruct* sock )
{
	int result;
	if( sock == NULL )
		return -1;
	result = connect( 
		sock->fd, (struct sockaddr *)&(sock->name), sizeof( sock->name ) );
	return result;
}

const char* sockConnectedTo( sockStruct* sock )
{
	static struct hostent* host;
	host = 
		gethostbyaddr( (char*)&sock->name.sin_addr, 
						sizeof( struct in_addr ), AF_INET );
	return host->h_name;
}
		
int sockSend( sockStruct* sock, void* data, size_t size )
{
	if( sock == NULL )
		return -1;
	return
		sendto( sock->fd, data, size, 0, (struct sockaddr*)&sock->name, sizeof( sock->name ) );
}

int sockRecv( sockStruct* sock, sockStruct* result, void* data, size_t size )
{
	int fromlen;
	int nbytes;
		
	if( sock == NULL || result == NULL )
		return -1;
	fromlen = sizeof( sock->name );

	nbytes = 
		recvfrom( sock->fd, data, size, 0, (struct sockaddr*)&result->name, &fromlen );

	result->fd = sock->fd;
	result->prot = sock->prot;
	result->port = ntohs( result->name.sin_port );
	return nbytes;
}

const char* sockPeerName( int fd )
{
	static struct hostent* host;
	struct sockaddr_in name;
	int size;
	size = sizeof( name );
	
	if( getpeername( fd, (struct sockaddr*)&name, &size ) < 0 )
		return NULL;
	host = 
		gethostbyaddr( (char*)&name.sin_addr, 
						sizeof( struct in_addr ), AF_INET );
	return host->h_name;
}
/* EOF libtcpip.c */
