/**
 ** File:   ttyfuncs.c
 ** Date:   Sun Oct 15 1995
 ** Author: Merijn de Jonge
 ** Email:  mdejonge@fwi.uva.nl
 **
 ** This file is part of the modem communication package.
 **
 ** Copyright (C) 1995 Merijn de Jonge.
 **
 ** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 ** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 ** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 ** 
 ** Redistribution and use in source and binary forms are permitted provided
 ** that: (1) source distributions retain this entire copyright notice and
 ** comment, and (2) distributions may not be sold for profit on physical
 ** media such as disks, tapes, and CD-ROMS, without expressed written
 ** permission.
 ** 
 **/
/* 
 * Parts of this file are based on dip by Fred N. van Kempen.
 *
 *
 * dip      A program for handling dialup IP connecions.
 *    UNIX terminal I/O support functions.  This file takes
 *    care of opening, setting up and maintaining the line,
 *    and it takes care of allocating the buffers and init-
 *    ializing their control structures.
 *
 * Version: @(#)tty.c   3.3.6 12/13/93
 *
 * Author:      Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *    Copyright 1988-1993 MicroWalt Corporation
 *    Lock file stuff stolen from Taylor UUCP
 *    Copyright (C) 1991, 1992 Ian Lance Taylor
 * 
 *    This program is free software; you can redistribute it
 *    and/or  modify it under  the terms of  the GNU General
 *    Public  License as  published  by  the  Free  Software
 *    Foundation;  either  version 2 of the License, or  (at
 *    your option) any later version.
 */   

#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#ifdef LINUX
#   include <linux/fs.h>
#   include <linux/tty.h>
#ifndef NO_SERIAL
#   include <linux/serial.h> /* for Linux-1.1.13 */
#endif /* ~NO_SERIAL */
#endif

#include "ttyfuncs.h"

typedef struct
{
	int fd;
	char* dev;
	int	locked;
	struct termios	tty_saved;
	struct termios	tty_current;
} _ttyData;

static struct {
  char   *speed;
  int code;
} tty_speeds[] = {         /* table of usable baud rates */
  { "0",        B0      },
  { "50",   B50   }, { "75",  B75   },
  { "110",  B110  }, { "300", B300  }, 
  { "600",  B600  }, { "1200",   B1200 },
  { "2400", B2400 }, { "4800",   B4800 },
  { "9600", B9600 },
#ifdef B14400
  { "14400",   B14400   },
#endif
#ifdef B19200
  { "19200",   B19200   },
#endif
#ifdef B38400
  { "38400",   B38400   },
#endif
#ifdef B57600
  { "57600",   B57600   },
#endif
#ifdef B115200
  { "115200",  B115200  },
#endif
  { NULL,   0  }
};


static _ttyData* ttyData = NULL;

char* tty_device_name()
{
	return ttyData->dev;
}

static _ttyData* tty_alloc()
{
	_ttyData* tty = (_ttyData*)malloc( sizeof( _ttyData ) );
	tty->fd = -1;
	tty->dev = NULL;
	tty->locked = 0;
	return tty;
}

static void tty_free( _ttyData* tty )
{
	free( tty->dev );
	free( tty );
}

int
tty_already_locked(char *dev) /* nam - complete path to lock file */
{
  int  i = 0, pid = 0;
  FILE *fd = NULL;
	char* lock_file = lock_file_name( dev );
  /* Does the lock file on our device exist? */
	fd = fopen( lock_file, "r" );
	if( fd == NULL )
		return 0;

  /* Yes, the lock is there.  Now let's make sure */
  /* at least there's no active process that owns */
  /* that lock.                                   */
  i = fscanf(fd, "%d", &pid);
  fclose(fd);
  
  if (i != 1) /* Lock file format's wrong! Kill't */
    return 0;

  /* We got the pid, check if the process's alive */
  if (kill(pid, 0) == 0)      /* it found process */
      return 1;           /* Yup, it's running... */

  /* Dead, we can proceed locking this device...  */
  return 0;
}

static int
tty_lock( _ttyData* tty, int mode)
{
   static char* lock_file;
   struct passwd *pw;
   pid_t ime;
   FILE *fd; 
#if HAVE_V2_LOCKFILES
   int cwrote;
#endif /* HAVE_V2_LOCKFILES */
   lock_file = lock_file_name( tty->dev );

   if (mode == 1)
   {  
      /* lock */
      if (tty_already_locked( tty->dev ) == 1)
      {
#ifdef DEBUG
         error( "libtty: attempt to use already locked tty %s",
            lock_file);
#endif
         return -1;
      }
      if ((fd = fopen(lock_file, "w")) == (FILE *)0)
      {
#ifdef DEBUG
         error( "libtty: lock: (%s): %m", lock_file);
#endif
         return -1;
      }
      ime = getpid();
#ifdef HAVE_V2_LOCKFILES
      cwrote = write (fileno(fd), &ime, sizeof(ime));
#else
      fprintf(fd, "%10d\n", (int)ime);
#endif
      fclose(fd);
       
      /* Make sure UUCP owns the lockfile.  Required by some packages. */
      if ((pw = getpwnam(_UID_UUCP)) == NULL) 
      {
#ifdef DEBUG
         error( "libtty: lock: UUCP user %s unknown!", _UID_UUCP);
#endif
         return 0;  /* keep the lock anyway */
         }
       chown(lock_file, pw->pw_uid, pw->pw_gid);
      tty->locked = 1;
   } 
   else
   if (mode == 2)
   { 
   	/* re-acquire a lock after a fork() */
      if (tty->locked != 1) 
      {
#ifdef DEBUG
      	error( "libtty: tty_lock reaquire: lock was not saved!");
#endif
      	return -1;
      }
      if ((fd = fopen(lock_file, "w")) == (FILE *)0)
      {
#ifdef DEBUG
      	error( "libtty: tty_lock: reacquire (%s): %m", lock_file );
#endif
      	return -1;
      }
      ime = getpid();
#ifdef HAVE_V2_LOCKFILES
      cwrote = write (fileno(fd), &ime, sizeof(ime));
#else
      fprintf(fd, "%10d\n", (int)ime);
#endif
      fclose(fd);
      chmod(lock_file, 0444);
      chown(lock_file, getuid(), getgid());
      return 0;
	} 
   else
   {   
   	/* unlock */
      if ( tty->locked != 1)
      {
#ifdef DEBUG
      	error( "libtty: tty_lock: lock was not saved?!");
#endif
         return 0;
		}
      if ((fd = fopen(lock_file, "w")) == (FILE *)0)
      {
#ifdef DEBUG
      	error( "libtty: tty_lock: can't reopen to delete: %m" );
#endif
			return -1;
		}
      if (unlink(lock_file) < 0)
      {
#ifdef DEBUG
			error( "libtty: unlock: (%s): %m", lock_file );
#endif
			tty->locked = 0;
			return -1;
		}
      tty->locked = 0;
	}
	return 0;
}

/* Find a serial speed code in the table. */
static int
tty_find_speed(char *speed)
{
  int i;

  i = 0;
  while (tty_speeds[i].speed != NULL) {
   if (! strcmp(tty_speeds[i].speed, speed)) return(tty_speeds[i].code);
   i++;
  }
  return -EINVAL;
}

/* Set the number of stop bits. */
static int
tty_set_stopbits(struct termios *tty, char *stopbits)
{
#ifdef DEBUG
   fprintf( stderr, "libtty: set_stopbits: %c", *stopbits);
#endif
  switch(*stopbits) {
   case '1':
      tty->c_cflag &= ~CSTOPB;
      break;

   case '2':
      tty->c_cflag |= CSTOPB;
      break;

   default:
      return -EINVAL;
  }
  return 0;
}

/* Set the number of data bits. */
static int
tty_set_databits(struct termios *tty, char *databits)
{
#ifdef DEBUG 
   fprintf(stderr, "libtty: set_databits: %c", *databits);
#endif
  tty->c_cflag &= ~CSIZE;
  switch(*databits) {
   case '5':
      tty->c_cflag |= CS5;
      break;

   case '6':
      tty->c_cflag |= CS6;
      break;

   case '7':
      tty->c_cflag |= CS7;
      break;

   case '8':
      tty->c_cflag |= CS8;
      break;

   default: 
      return -EINVAL;
  }
  return 0;
}  

/* Set the type of parity encoding. */
static int
tty_set_parity(struct termios *tty, char *parity)
{
#ifdef DEBUG
   fprintf( stderr, "libtty: set_parity: %c", *parity);
#endif
  switch(toupper(*parity)) {
   case 'N':
      tty->c_cflag &= ~(PARENB | PARODD);
      break;  

   case 'O':
      tty->c_cflag &= ~(PARENB | PARODD);
      tty->c_cflag |= (PARENB | PARODD); 
      break;

   case 'E':
      tty->c_cflag &= ~(PARENB | PARODD);
      tty->c_cflag |= (PARENB);
      break;

   default:
      return -EINVAL;
  }
  return 0;
}

/* Set the line speed of a terminal line. */
static int
tty_set_speed(struct termios *tty, char *speed)
{
	int code;

#ifdef DEBUG
	printf("libtty: set_speed: %s", speed);
#endif
	code = tty_find_speed(speed);
	if (code < 0) 
		return -1;
	tty->c_cflag &= ~CBAUD;  
	tty->c_cflag |= code;    

	return 0;
}
 
/* Put a terminal line in a transparent state. */
int
tty_set_raw(struct termios *tty)
{
	int i;
	int speed;

	for(i = 0; i < NCCS; i++)
		tty->c_cc[i] = '\0';

	tty->c_cc[VMIN] = 1;
	tty->c_cc[VTIME] = 0;
	tty->c_iflag = IGNBRK | IGNPAR;
	tty->c_oflag = 0;
	tty->c_lflag = 0;
	speed = tty->c_cflag & CBAUD; 
	tty->c_cflag = CRTSCTS|HUPCL|CREAD|CLOCAL;
	tty->c_cflag |= speed;

	return 0;
/*
	tty->c_iflag &= ~(IGNBRK | IGNCR | INLCR | ICRNL | IUCLC | 
		IXANY | IXON | IXOFF | INPCK | ISTRIP);
	tty->c_iflag |= (BRKINT | IGNPAR);
	tty->c_oflag &= ~OPOST;
	tty->c_lflag &= ~(XCASE|ECHONL|NOFLSH);
	tty->c_lflag &= ~(ICANON | ISIG | ECHO);
	tty->c_cflag |= CREAD;
   
*/	
	return 0;
}

/* Fetch the state of a terminal. */
static int
tty_get_state( int fd, struct termios* tty )
{
	if (fd >= 0) 
	{
		if (ioctl(fd, TCGETS, tty ) < 0) 
		{
#ifdef DEBUG
			error("libtty: get_state: %m" );
#endif
			return -errno;
		}
	   return 0;
	}
	return -1;
}
 
/* Set the state of a terminal. */
static int
tty_set_state( int fd, struct termios *tty)
{
	if (fd >= 0) 
	{
		if (ioctl(fd, TCSETS, tty) < 0) 
		{
#ifdef DEBUG
			error( "libtty: set_state: %m" );
#endif
			return -errno;
		}
		return 0;
	}
	return -1; 
}  

/* Return the TTY link's file descriptor. */
int
tty_askfd( )
{  
	return ttyData->fd;
}

/* Set the number of databits a terminal line. */
int
tty_databits(char *bits)
{  
	if (tty_set_databits(&ttyData->tty_current, bits) < 0) 
		return -1;
	return tty_set_state(ttyData->fd, &ttyData->tty_current);
}

/* Set the number of stopbits of a terminal line. */
int
tty_stopbits(char *bits)
{  
	if (tty_set_stopbits(&ttyData->tty_current, bits) < 0) 
		return -1;
	return tty_set_state(ttyData->fd, &ttyData->tty_current);
}

/* Set the type of parity of a terminal line. */
int
tty_parity(char *type)
{  
	if (tty_set_parity(&ttyData->tty_current, type) < 0) 
		return -1;
	return tty_set_state( ttyData->fd, &ttyData->tty_current);
}	

/* Set the line speed of a terminal line. */
int
tty_speed(char *speed)
{
#if defined(LINUX)
	struct serial_struct info;
	int spd;
	
	spd = atoi(speed);
	if (ioctl( ttyData->fd, TIOCGSERIAL, &info)) 
	{
#ifdef DEBUG
		error( "Error ioctl1 %m");
#endif
		return -1;
	}
	
	info.flags = info.flags & ~ASYNC_SPD_MASK;

	if (spd == 115200) 
	{
		if (tty_set_speed(&ttyData->tty_current, "38400") < 0) 
			return -1;
		info.flags |= ASYNC_SPD_VHI;
	} 
	else 
	if (spd == 57600) 
	{
		if (tty_set_speed(&ttyData->tty_current, "38400") < 0) 
			return -1;
		info.flags |= ASYNC_SPD_HI;
	} 
	else 
	{
		if (tty_set_speed(&ttyData->tty_current, speed) < 0) 
			return -1;
	}
	if (ioctl(ttyData->fd, TIOCSSERIAL, &info)) 
	{
#ifdef DEBUG
		error( "Error ioctl 2 %,");
#endif
		return -1;
	}
#else
	if (tty_set_speed(&ttyData->tty_current, speed) < 0) 
		return -1;
#endif

	return tty_set_state( ttyData->fd, &ttyData->tty_current);
}

/* Hangup the line. */
int
hanguptty(void)
{
	struct termios tty;
	struct termios tty_current;
	tty_get_state( ttyData->fd, &tty_current );
	
	
	tty = tty_current;
	tty_set_speed(&tty, "0");
	if (tty_set_state( ttyData->fd, &tty) < 0) 
	{
#ifdef DEBUG
		error( "libtty: hangup(DROP): %m" );
#endif
		return -errno;
	}

	sleep(3);
  
	if (tty_set_state( ttyData->fd, &tty_current) < 0) 
	{
#ifdef DEBUG
		error( "libtty: hangup(RAISE): %m" );
#endif
		return -errno;
	}
	return 0;
}
   
/* Clear the CLOCAL bit. (detect carrier loss) */
int
tty_notlocal(void)
{
	if (ttyData->fd < 0)
	return -1;  

	/* Release any control terminal we might have. */
	if (setsid() < 0) 
	{
#ifdef DEBUG
	error( "libtty: tty_notlocal cannot setsid: %m" );
#endif
	}   
      
	if (ioctl(ttyData->fd, TIOCSCTTY, 1) < 0) 
	{
#ifdef DEBUG
		error("libtty: tty_notlocal cannot TIOCSCTTY: %m");
#endif
	}   
      
	ttyData->tty_current.c_cflag &= ~CLOCAL;
	if (tty_set_state( ttyData->fd, &ttyData->tty_current) < 0) 
	{
#ifdef DEBUG
		error("libtty: tty_notlocal cannot clr CLOCAL: %m" );
#endif
		return(-errno);
	}
	return 0;
}

/* Flush input on the terminal. */
int
tty_flush( int fd)
{
	return(0);
}
 
 
/* Close down a terminal line. */
int
tty_close(void)
{
	if (ttyData->fd < 0)
		return -1;

	if (tty_set_state( ttyData->fd, &ttyData->tty_saved) < 0) 
	{
#ifdef DEBUG
		error( "libtty: tty_close: restore: %m" );
#endif
	   close(ttyData->fd);
   	ttyData->fd = -1;
		sleep (1);
	   tty_lock( ttyData, 0);
		tty_free( ttyData );
		ttyData = NULL;
	   return -errno;
	}
   
	hanguptty();
	close(ttyData->fd);
	ttyData->fd = -1;
	tty_lock( ttyData, 0);
	tty_free( ttyData );
	ttyData = NULL;
		
  return 0;
}

/* Open and initialize a terminal line. */
int
tty_open(char *name)
{
	int flags;
	char buf[PATH_MAX];
	ttyData = tty_alloc();

  /* Try opening the TTY device. */
	if (name != NULL)
	{
		ttyData->dev = strdup(  make_path( name ) );
	   /* Now - can we lock it? */
	   if(tty_lock( ttyData, 1)) 
	   {
#ifdef DEBUG
      	error( "libtty: can't open - problems with locking %s",
				ttyData->dev);
#endif
	      return -1;
   	}
		sprintf( buf, "/dev/%s", ttyData->dev );

		ttyData->fd = open(buf, O_RDWR|O_NDELAY);
	   if( ttyData->fd < 0 ) 
	   {
#ifdef DEBUG
      	error( "libtty: open(%s, RW): %m", buf);
#endif
      	tty_lock( ttyData, 0);
	      return -errno;
   	}
#ifdef DEBUG   
		error( "libtty: tty_open: %s (%d) ", buf, ttyData->fd);
#endif
	} 
	else 
		ttyData->fd = 0;
   
   
	/* Fetch the current state of the terminal. */
	if (tty_get_state( ttyData->fd, &ttyData->tty_saved) < 0) 
	{
#ifdef DEBUG
   	error( "libtty: tty_open: cannot get current state!");
#endif
	   return -errno;
  	}
   
	memcpy ((char *)&ttyData->tty_current,
		(char *)&ttyData->tty_saved, sizeof(struct termios));

	/* Put this terminal line in a 8-bit transparent mode. */
	if (tty_set_raw(&ttyData->tty_current) < 0) 
	{
#ifdef DEBUG
		error( "libtty: tty_open: cannot set RAW mode!");
#endif
		return -errno;
	}   

	/* If we are running in MASTER mode, set the default speed. */
	if ((name != NULL) && (tty_set_speed(&ttyData->tty_current, "38400") != 0)) 
	{
#ifdef DEBUG
   	error( "libtty: tty_open: cannot set 38400 bps!");
#endif
		return -errno;
	}   

  /* Set up a completely 8-bit clean line. */
	if (tty_set_databits(&ttyData->tty_current, "8") ||
		 tty_set_stopbits(&ttyData->tty_current, "1") || 
		 tty_set_parity(&ttyData->tty_current, "N")) 
	{  
#ifdef DEBUG
		error( "libtty: tty_open: cannot set 8N1 mode!");
#endif
		return -errno;
	}   

	/* Set the new line mode. */
	if( tty_set_state( ttyData->fd, &ttyData->tty_current) < 0) 
		return -errno;
	/* Clear the NDELAY flag now (line is in CLOCAL) */
	flags = fcntl( ttyData->fd, F_GETFL, 0);
	fcntl( ttyData->fd, F_SETFL, flags & ~O_NDELAY);

  /* OK, all done.  Lock this terminal line. */
	return ttyData->fd;
}

void
tty_sendbreak(void)
{
	if (ttyData->fd >= 0)
		tcsendbreak(ttyData->fd, 0);
#if 0
	ioctl( ttyData->fd, TCSBRK, 0);
#endif
}
 
char* make_path( char* name )
{
	register char *sp;

	if (name != NULL) 
	{
		sp = strrchr( name, '/' );
		if( sp != NULL )
			return sp+1;
		else
			return name;
	}
	return NULL;
}

char* lock_file_name( char* dev )
{
	static char lock_file[PATH_MAX];
   char* ptr;
   ptr = strrchr( dev , '/' );
   if( ptr != NULL )
      *ptr++ = '\0';
   else
      ptr = dev;
   sprintf( lock_file, "%s/LCK..%s", _PATH_LOCKD, ptr );
   return lock_file;
}
/* EOF ttyfuncs.c */
