/**
 ** File:   client.c
 ** Date:   Sun Oct 15 1995
 ** Author: Merijn de Jonge
 ** Email:  mdejonge@fwi.uva.nl
 **
 ** This file is part of the modem communication package.
 **
 ** Copyright (C) 1995 Merijn de Jonge.
 **
 ** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 ** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 ** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 ** 
 ** Redistribution and use in source and binary forms are permitted provided
 ** that: (1) source distributions retain this entire copyright notice and
 ** comment, and (2) distributions may not be sold for profit on physical
 ** media such as disks, tapes, and CD-ROMS, without expressed written
 ** permission.
 ** 
 **/
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>

#include <libtty/libtty.h>
#include <libtcpip/libtcpip.h>
#include <libmodem/libmodem.h>
#include <libconf/libconf.h>
#include <liberror/liberror.h>
#include <libprotocol/libprotocol.h>
#include "log.h"

static modem_commands commands =
{
	ATTN,
	RSET,
	INIT,
	DIAL,
	HANGUP
};

static char buf[512];
static char* remoteHost;
static char* remoteUser;
static char  localHost[40];

static void do_transfer();
static void client_exit();
static int sock;
static int connected = 0;
static time_t connect_since;

static void handler( int sig )
{	
	signal( sig, SIG_IGN );
	switch( sig )
	{
		case SIGHUP:
			error( "Hangup" );
			break;
	}
	signal( sig, handler );
}

static void clientErrorFunc( const char* err )
{
	protSendError( sock, err, strlen( err ) + 1 );
	exit( 1 );
}

static void setup( char* mdm )
{
	conf_file* cf;
	char mdm_type[20];

	strcpy( mdm_type, DEF_MODEM_TYPE );
	
	cf = open_conf( MODEM_CONF_FILE );
	if( cf != NULL )
	{
		strcpy( mdm_type, get_str( cf, mdm, "type", DEF_MODEM_TYPE ) );
		close_conf(cf );
	}
	cf = open_conf( MODEM_COMMAND );
	if( cf != NULL )
	{
		strcpy( commands.AT_ATTN, get_str( cf, mdm_type, "attn", ATTN ) );	 
		strcpy( commands.AT_RSET, get_str( cf, mdm_type, "rset", RSET ) );	 
		strcpy( commands.AT_INIT, get_str( cf, mdm_type, "init", INIT ) );	 
		strcpy( commands.AT_DIAL, get_str( cf, mdm_type, "dial", DIAL ) );	 
		strcpy( commands.AT_HANGUP, get_str( cf, mdm_type, "hangup", HANGUP ) );	 
		close_conf( cf );
	}

#ifdef DEBUG
   error( "Modem commands:" );
   error( "AT_ATTN = %s", commands.AT_ATTN );
   error( "AT_RSET = %s", commands.AT_RSET );
   error( "AT_INIT = %s", commands.AT_INIT );
   error( "AT_DIAL = %s", commands.AT_DIAL );
   error( "AT_HANGUP = %s", commands.AT_HANGUP );
#endif

	protSend( sock, &commands, sizeof(commands ) );
}


void handle_client( int _sock, const char* host, const char* uname )
{
	pid_t pid;

	pid = fork();
	switch( pid )
	{
		case 0:
			atexit( client_exit );
			remoteHost = strdup( host );
			remoteUser = strdup( uname );
			gethostname( localHost, sizeof(localHost) );
			sock = _sock;
			fprintf( ACCESS_LOG, "%s[%s] Connect from: \"%s@%s\"\n",
				timeStamp(),
				localHost,
				remoteUser,
				remoteHost );
			fflush( ACCESS_LOG );
			do_transfer();
			exit( 0 );
		case -1:
			error( "[fork] %m" );
			exit( 1 );
		default:
			return;
	}
}

static void do_transfer( )
{
	int tty;
	fd_set fds;
	fd_set orig;
	int retval;
	int bytes;
	int status;
	unsigned int arg;	
	struct timeval timeout;
	
	setErrorFunc( clientErrorFunc );

	tty = ttyOpen();
	if( tty == -1 )
		error( "No modems available" );

	setup( ttyName() );
	

	FD_ZERO( &orig	);
	FD_SET( sock, &orig );
	FD_SET( tty, &orig );
	
	
	while( 1 )
	{
		fds = orig;
		timeout.tv_sec = CLIENT_TIMEOUT;
		timeout.tv_usec = 0;
		
		retval = select( FD_SETSIZE, &fds, NULL, NULL, &timeout );
		if( retval < 0 )
			error( "[select] %m" );

		if( retval == 0 )
		{
			if( !connected )
				error( "You didn't do anything, bye" );
			continue;
		}
		
		if( FD_ISSET( sock, &fds ) )
		{
			bytes = protReceive( sock, buf, sizeof(buf ) );
			if( bytes <= 0 )
			{
				if( errno == EINTR )
					continue;
				if( bytes == -1 )
					error( "[read sock] %m");
				exit( 0 );
			}
			bytes = write( tty, buf, bytes );
			if( bytes <= 0 )
			{
				if( errno == EINTR )
					continue;
				if( errno == EIO )
					error( "Hangup" );
				if( bytes < 0 )
					error( "[write sock] %m" );
				exit( 0 );
			}

#ifdef debug
			buf[bytes] = '\0';
			error( "SOCK: %s", buf );
#endif
		}

		if( FD_ISSET( tty, &fds ) )
		{
		
			if( !connected )
			{
				status = ioctl( tty, TIOCMGET, &arg );
				if(  arg &TIOCM_CAR )
				{
					time( &connect_since );
					connected = 1;
				}
			}	
			bytes = read( tty, buf, sizeof( buf ) );
			if( bytes <= 0 )
			{
				if( errno == EINTR )
					continue;
				if( errno == EIO )
					error( "Hangup" );
				if( bytes < 0 )
					error( "[read tty] %m");
				exit( 0 );
			}
			bytes = protSend( sock, buf, bytes );
			if( bytes <= 0 )
			{
				if( errno == EINTR )
					continue;
				if( bytes == -1 )
					error( "[write sock] %m" );
				exit( 0 );
			}			
				
#ifdef DEBUG
			buf[bytes] = '\0';
			error( "MODEM: %s", buf );
#endif
		}
	}
}

		
void client_exit()
{
	char buf[100];
	time_t now;
	struct tm* tm;
	
	buf[0] = '\0';
	if( connected )
	{
		time( &now );
		now = difftime( now, connect_since );
		tm = gmtime( &now );
		strftime( buf, sizeof(buf), "%H:%M:%S", tm );
		fprintf( ACCOUNT_LOG, "%s[%s] \"%s@%s\" using \"%s\" time %s\n",
			timeStamp(),
			localHost,
			remoteUser,
			remoteHost,
			ttyDev(),
			buf );
		fflush( ACCOUNT_LOG );
	}
	fprintf( ACCESS_LOG, "%s[%s] Disconnect from: \"%s@%s\"\n",
		timeStamp(),
		localHost,
		remoteUser,
		remoteHost );
	fflush( ACCESS_LOG );

	close( sock );
	ttyClose();
	closeLog();
}
/* EOF client.c */
