/**
 ** File:   daemon.c
 ** Date:   Sun Oct 15 1995
 ** Author: Merijn de Jonge
 ** Email:  mdejonge@fwi.uva.nl
 **
 ** This file is part of the modem communication package.
 **
 ** Copyright (C) 1995 Merijn de Jonge.
 **
 ** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 ** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 ** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 ** 
 ** Redistribution and use in source and binary forms are permitted provided
 ** that: (1) source distributions retain this entire copyright notice and
 ** comment, and (2) distributions may not be sold for profit on physical
 ** media such as disks, tapes, and CD-ROMS, without expressed written
 ** permission.
 ** 
 **/
#include <sys/ioctl.h>
#include <sys/file.h>    
#include <sys/stat.h> /* for umask */
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <liberror/liberror.h>
#include <modem_defs.h>
#include "log.h"

static int Fork( void );
static int CloseFileDescriptors( void );
static int DetachTTY( void );
static int Chdir( char* serverDir );
static int Umask( int Umask );
static int OpenDummys( void );
static int LockFile( char* lockFile );

/*
 * Function: startup_daemon
 *
 * Description:
 *   Create daemon process by forking, and let parent exit
 *
 * Arguments:
 *   - int   daemon   Indicates whether or not we go in
 *                     background by forking.
 * Returns:
 *   - int   0     On success
 *
 *
 * Steps taken in this function:
 *
 * 1) fork, and let parent exit  (if daemon == 0)
 * 2) detach from controlling tty (if daemon == 0 )
 * 3) Chdir to known directory ('directory' in config file)
 *    default: SERVER_DIR
 * 4) Set umask ( 'umask' in config file )
 *    default: UMASK
 * 5) Create lock file ('lock file'  in conf file)
 *    default: LOCKFILE
 * 6) Close all file descriptors (if daemon == 0 )
 * 7) Open dummy file descriptors (if daemon == 0 )
 *    (/dev/null) for stdin, stdout
 *
 */
int startup_daemon( int daemon)
{
   int    umask;
   char* serverDir;
   char* lockFile;
   
   umask     = SERVER_UMASK;
   serverDir = SERVER_DIR;
   lockFile  = SERVER_LOCK_FILE;

   /* step 1 */
   if( daemon )
		Fork();
   
   /* step 2 */
   if( daemon )
   	DetachTTY();
   
   /* step 3 */
   Chdir( serverDir );
   
   /* step 4 */
   Umask( umask );
   
   
   /* step 5 */
   LockFile( lockFile );
   
   /* step 6 */
   if( daemon )
   	CloseFileDescriptors();
   /* step 7 */
   if( daemon )
   	OpenDummys();
   
   return 0;
}


/*
 * Function: Fork
 *
 * Description:
 *   Step 1 in startup_daemon function.
 *   fork and let parent exit.
 *
 * Arguments:
 *   - None
 *
 * Returns:
 *   - int   0   On success
 *
 * Remarks:
 *   This function exits on error
 *
 *
 */
static int Fork( void )
{
   switch( fork() )
   {
      case 0: /* child */
         break;

      case -1:
         error( "fork: %m" ); 
         exit( 1 );
      default: /* parent */
         exit(0);   /* parent exists*/
   }
   return 0;
}

/*
 * Function: CloseFileDescriptors
 *
 * Description:
 *   Step 6 in startup_daemon function.
 *   close all filedescriptors of current process.
 *
 * Arguments:
 *   - None
 *
 * Returns:
 *   - int   0   On success (this is always the case)
 *
 */
static int CloseFileDescriptors( void )
{
   int i;
   
   for( i = getdtablesize() - 1; i >= 0; --i )
      close( i );
   
   return 0;
}

/*
 * Function: DetachTTY
 *
 * Description:
 *   Step 2 in startup_daemon function.
 *   Detach from tty.
 *   Don't receive signals from process that started server
 *
 * Arguments:
 *   - None
 *
 * Returns:
 *   - int   0   On success (always the case)
 *
 */
static int DetachTTY( void )
{
   pid_t pgrp;
   pgrp = setsid();
   
	if( pgrp == -1 )
	{
		error( "setsid: %m" );
		exit( 1 );
	}
   return 0;
}

/*
 * Function: Chdir
 *
 * Description:
 *   Step 3 in startup_daemon function.
 *   Change to known directory where server runs.
 *
 * Arguments:
 *   - char*   serverDir   Dir. to chdir to.
 *
 * Returns:
 *   - int     0        On success (always the case)
 *
 * Remarks:
 *   This function exits on error
 *
 *
 */
static int Chdir( char* serverDir )
{
   if( chdir( serverDir ) < 0 )
   {
      error( "chdir: %m" );
      exit( 1 );
   }

   return 0;
}

/*
 * Function: Umask
 *
 * Description:
 *   Step 4 in startup_daemon function.
 *   Change umask of current process.
 *
 * Arguments:
 *   - int    umask   The umask to set.
 *
 * Returns:
 *   - int   0    On success (Always the case)
 *
 */
static int Umask( int Umask )
{
   umask( Umask );

   return 0;
}


/*
 * Function: OpenDummys
 *
 * Description:
 *   Step 7 in startup_daemon function.
 *   filedescr. 0, 1, 2 are closed. Since some standard
 *   functions get confused when one of these are closed,
 *   we use dummy's for them.
 *
 * Arguments:
 *   - None
 *
 * Returns:
 *   - int   0   On success (always the case)
 *
 */
static int OpenDummys( void )
{
   int fd;
   
   fd = open( "/dev/null", O_RDWR );   /* stdin */
	if( fd == -1 )
	{
		error( "open: \"%s\" %m", "/dev/null" );
		exit( 1 );
	}
   if( dup( fd ) == -1   ||  /* stdout */
       dup( fd ) == -1 )	  /* stderr */
   {
   	error( "dup: %m" );
   	exit( 1 );
   }
   return 0;
}

/*
 * Function: LockFile
 *
 * Description:
 *   Step 5 in startup_daemon function.
 *   Create lockfile, to avoid multiple copies,
 *   and write pid in lockfile.
 *
 * Arguments:
 *   - char*   lockfile   Filename of lockfile to use
 *
 * Returns:
 *   - int     0       On success (always the case)
 *
 * Remarks:
 *   This function exits on error.
 *
 */
static int LockFile( char* lockFile )
{
   int   lf;
   char    buf[10];
   
   lf = open( lockFile, O_RDWR | O_CREAT, 0640 );
   if( lf < 0 )      /* error opening file */
   {
      error( "open: \"%s\" %m", lockFile );
      exit( 1 );
   }
   if( flock( lf, LOCK_EX | LOCK_NB ) )
   {
      error( "server already started" );
      exit( 1 );    /* error with locking the file */
   }   
   
   /* Record process id. */
   sprintf( buf, "%-6d", (int)getpid() );
   write( lf, buf, strlen( buf ) );

   return 0;
}


/* EOF daemon.c */
