/**
 ** File:   server.c
 ** Date:   Sun Oct 15 1995
 ** Author: Merijn de Jonge
 ** Email:  mdejonge@fwi.uva.nl
 **
 ** This file is part of the modem communication package.
 **
 ** Copyright (C) 1995 Merijn de Jonge.
 **
 ** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 ** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 ** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 ** 
 ** Redistribution and use in source and binary forms are permitted provided
 ** that: (1) source distributions retain this entire copyright notice and
 ** comment, and (2) distributions may not be sold for profit on physical
 ** media such as disks, tapes, and CD-ROMS, without expressed written
 ** permission.
 ** 
 **/
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libtcpip/libtcpip.h>
#include <liberror/liberror.h>
#include <libprotocol/libprotocol.h>
#include <modem_defs.h>
#include "modemd.h"
#include "request.h"

static request_ID newId()
{
	time_t time_stamp;
	time( &time_stamp );
	return (request_ID)time_stamp;
}

static int verify_user( const char* host,  const char* uname )
{
	pid_t pid;
	int status;
	
	pid = fork();

	switch( pid )
	{
		case -1:
			error( "fork: %m" );
			exit( 1 );
		case 0:
			execl( VERIFY_USER, VERIFY_USER, uname, host, NULL );
			error( "execl: %m" );
			exit( 1 );
		default:
			waitpid( pid, &status, 0 );
			if( WIFEXITED( status ) )
				return ( WEXITSTATUS( status ) == 1 );
	}	
	return 0;
}

static int check_client( int sock, request* clientInfo )
{
	request_ID id;
	int ind;
	request* req;
	char* err_msg = "Go away";
	
	if( protReceive( sock, &id, sizeof( id ) ) < 0 )
	{
		error( "read: %m" );
		exit( 1 );
	}
	ind = requestListLookup( id );
	if( ind == -1 )
	{
		write( sock, err_msg, strlen( err_msg ) + 1 );
		return 0;
	}

	req = requestListGet( id );
	strcpy( clientInfo->uname, req->uname );
	strcpy( clientInfo->host, req->host );
	requestListRemove( ind );
	return 1;
}
	
	
static void new_request( sockStruct* sock )
{
	sockStruct client;
	int nbytes;
	request_ID id;
	request* req;
			
	req = (request*)malloc( sizeof( request ) );
	
	if( req == NULL )
	{
		error( "malloc: not enough memory" );
		exit( 1 );
	}

	nbytes = sockRecv( sock, &client, req->uname, sizeof( req->uname ) );
	if( nbytes < 0 )
	{
		error( "recv: %m" );
		exit( 1 );
	}
	strncpy( req->host, sockConnectedTo( &client ), sizeof( req->host ) );
	
	if( verify_user( req->host, req->uname ) == 0 )
	{
		free( req );
		return;
	}
	
	id = newId();
	
	req->id = id;
	
	if( requestListAdd( req ) == -1 )
	{
		free( req );
		return;
	}
			
	nbytes = sockSend( &client, &id, sizeof( id ) );
}

	
void server()
{
	sockStruct sock_tcp;
	sockStruct sock_udp;
	sockStruct client_tcp;
		
	fd_set active_fd_set, read_fd_set;
	int i;
	int retval;
	struct timeval timeout;
	
	requestListInit();
	
	sockSetProtocol( &sock_tcp, TCP );
	sockSetProtocol( &sock_udp, UDP );
	
	if( sockSetPortToService( &sock_tcp, TCP_SERVICE ) == -1 )
	{
		error( "Unknown service: %s, using default port: %d", 
			TCP_SERVICE, TCP_DEFAULT_PORT );
		sockSetPort( &sock_tcp, TCP_DEFAULT_PORT );
	}
	if( sockSetPortToService( &sock_udp, UDP_SERVICE ) == -1 )
	{
		error( "Unknown service: %s, using default port: %d", 
			UDP_SERVICE, UDP_DEFAULT_PORT );
		sockSetPort( &sock_udp, UDP_DEFAULT_PORT );
	}

	if( 
		sockServer( &sock_tcp ) ||
		sockListen( &sock_tcp ) ||
		sockServer( &sock_udp )
		== -1 )
	{
		error( "setup error: %m" );
		exit( 1 );
	}
	
	FD_ZERO( &active_fd_set );
	FD_SET( sock_tcp.fd, &active_fd_set );
	FD_SET( sock_udp.fd, &active_fd_set );
	
	while( 1 )
	{
		read_fd_set = active_fd_set;
		timeout.tv_sec = 5;
		timeout.tv_usec = 0;
		
		retval= select( FD_SETSIZE, &read_fd_set, NULL, NULL, &timeout );
		if( retval < 0 )
		{
			error( "select: %m" );
			exit( 1 );
		}
		if( retval == 0 )
		{
			requestListReset();
			continue;
		}
		for( i = 0; i < FD_SETSIZE; ++i )
		if( FD_ISSET( i, &read_fd_set ) )
		{
			if( i == sock_tcp.fd )
			{
				if( sockAccept( &sock_tcp, &client_tcp ) == -1 )
				{
					error( "sockAccept: %m" );
					exit( 1 );
				}
				FD_SET( client_tcp.fd, &active_fd_set );
			}
			else
			if( i == sock_udp.fd )
			{
				new_request( &sock_udp );
			}
			else
			{
				request req;
				if( check_client( i, &req ) )
				{
					handle_client( i, req.host, req.uname );
				}
				close( i );
				FD_CLR( i, &active_fd_set );
			}
		}
	}
}
/* EOF server.c */
