/**
 ** File:   main.c
 ** Date:   Sun Oct 15 1995
 ** Author: Merijn de Jonge
 ** Email:  mdejonge@fwi.uva.nl
 **
 ** This file is part of the modem communication package.
 **
 ** Copyright (C) 1995 Merijn de Jonge.
 **
 ** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 ** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 ** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 ** 
 ** Redistribution and use in source and binary forms are permitted provided
 ** that: (1) source distributions retain this entire copyright notice and
 ** comment, and (2) distributions may not be sold for profit on physical
 ** media such as disks, tapes, and CD-ROMS, without expressed written
 ** permission.
 ** 
 **/
#include <stdio.h>
#include <setjmp.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <signal.h>

#include <libdialogs/libdialogs.h>
#include <libphonebook/libphonebook.h>
#include <libtty/libtty.h>
#include <libmodem/libmodem.h>
#include <libprotocol/libprotocol.h>
#include <libtcpip/libtcpip.h>
#include "jobc.h"
#include "modem_reply.h"
#include "connect.h"


static jmp_buf env;
static int modemDevice;
static int ptyMasterDev;
static int auto_retry = 0;
static modem_commands commands;
static int interface = 0;

static char* localProgName = NULL;
static char* phonenumber = NULL;
static struct option long_options[] =
{
	{ "use",     1, 0, 'u' },
	{ "program", 1, 0, 'p' },
	{ "auto",    0, 0, 'a' },
	{ 0,         0, 0,  0  }
};

static char* opt_string = "u:U:p:P:aA";


void test()
{
	fprintf( stderr, "Modem commands:\n" );
	fprintf( stderr, "AT_ATTN = %s\n", commands.AT_ATTN );
	fprintf( stderr, "AT_RSET = %s\n", commands.AT_RSET );
	fprintf( stderr, "AT_INIT = %s\n", commands.AT_INIT );
	fprintf( stderr, "AT_DIAL = %s\n", commands.AT_DIAL );
	fprintf( stderr, "AT_HANGUP = %s\n", commands.AT_HANGUP );
}


#define DO_DIAL
static void initStart( int argc, char* argv[] );
static void	initSetNumber( char* entry );
static void	initSetProgram( char* program );
static void getNumberStart( char* number );
static void getNumberCancel();
static void getNumberDone( char* number );
static void dialStart( char* number );
static void dialRetry();
static void dialCancel();
static void dialTimeout();
static void dialDo( char* );
static void dialError( char* );
static void dialDone( char* );
static void statusStart( char* );
static void signalHandler( int sig );

static void signalHandler( int sig )
{
	switch( sig )
	{
		case SIGPIPE:
			error( " \nBroken pipe,\n \nServer may have closed connection.\n" );
			break;
	}
}

static void xmmdm_error( const char* msg )
{
	if( interface )
		errorDialogPopup( msg );
	else
		fprintf( stderr, "ERROR: %s\n", msg );
	exit( 1 );
}

static void initStart( int argc, char* argv[] )
{
	int c;
	while( 1 )
	{
		int option_index = 0;
		c = getopt_long (argc, argv, opt_string, long_options, &option_index );
		if( c == -1 )
			break;
		switch( c )
		{
			case 'u':
			case 'U':
				initSetNumber( optarg );
				break;
			case 'p':
			case 'P':
				initSetProgram( optarg );
				break;
			case 'A':
			case 'a':
				auto_retry = 1;
				break;
			default:
				error( "Unrecognised option" );
				break;
		}
	}
}

static void initSetNumber( char* entry )
{
	phonebook_entry* e;
	phonebook* book;
	char* val;
	
	book = phonebookOpen( PHONEBOOK );
	if( book == NULL )
		return;
	e = phonebookGetEntryByName( book, entry );
	if( e != NULL )
	{
		val = numberValue( book, e );
		if( val != NULL )
			phonenumber = strdup( val );
	}

	phonebookClose( book );
}

static void initSetProgram( char* program )
{
	localProgName = strdup( program );
}

	
static void getNumberStart( char* number )
{
	dialDialogInfoRec info;
	info.number = number;
	info.cancelCallback = getNumberCancel;
	info.dialCallback   = getNumberDone;

	dialDialogInit( &info );
	dialDialogPopup();
	return;
}

static void getNumberCancel()
{
	exit( 0 );
}

static void getNumberDone( char* number )
{
	dialingDialogInfoRec info;
	info.number = number;
	info.timeout = 60;
	info.cancelCallback = dialCancel;
	info.timeoutCallback = dialTimeout;
	info.retryCallback = dialRetry;
	dialingDialogInit( &info );
	dialStart( number );
}

static void dialStart( char* number )
{
	int status;
	char buf[100];
	
	status = setjmp( env );
	switch( status )
	{
		case 0:
			dialDo( number );
			return;
		case 1:
			dialError( "Timeout" );
			break;
		case 2:
			mdm_hangup( modemDevice );
			modemRead( modemDevice, buf, sizeof(buf) );
			break;
		case 3:
			mdm_hangup( modemDevice );
			modemRead( modemDevice, buf, sizeof(buf) );
			dialingDialogReset();
			dialStart( number );
			return;
		default:
			error( "Wrong longjmp" );
			exit( 1 );
	}
	dialingDialogPopdown();
	getNumberStart( number );
}

static void dialRetry()
{
	
	longjmp( env, 3 );
}


static void dialCancel()
{
	longjmp( env, 2 );
}

static void dialTimeout()
{
	longjmp( env, 1 );
}

static void dialError( char* msg )
{
	char buf[100];
		
	sprintf( buf, "            %s            ", msg );
	errorDialogPopup( buf );
}

static void dialDo( char* number )
{
#ifdef DIAL
	int result;
	mdm_dial( modemDevice, number );
	dialingDialogPopup();
	result = expect( modemDevice, connectStr, MATCH_ALL_TEXT );
	if( result == 0 )
	{
		result = expect( modemDevice, speedStr, MATCH_ALL_TEXT );
		dialDone( (char*)speedStr[result] );
		return;
	}
	else
	{
		if( auto_retry )
			longjmp( env, 3 );
		dialingDialogPopdown();
		dialError( (char*)connectStr[result] );
		getNumberStart( number );
		return;
	}
#else
	dialDone( "28800" );
	return;
#endif
}
	
static void dialDone( char* bps )
{
	dialingDialogPopdown();
	statusStart( bps );
}

static int doRead( int fd )
{
	static char buf[512];
	int bytes = 0;
	
	if( fd == modemDevice )
	{
		bytes = modemRead( modemDevice, buf, sizeof( buf ) );
		if( bytes <= 0 )
		{
			error( "read modem: %m" );
			exit( 0 );
		}
		bytes = write( ptyMasterDev, buf, bytes );
		if( bytes <= 0 )
		{
			error( "write pty: %m" );
			exit( 1 );
		}
	}
	if( fd == ptyMasterDev )
	{
		bytes = read( ptyMasterDev, buf, sizeof( buf ) );
		if( bytes <= 0 )
		{
			error( "read pty: %m" );
			exit( 0 );
		}
		bytes = modemWrite( modemDevice, buf, bytes );
		if( bytes <= 0 )
		{
			error( "write modem: %m" );
			exit( 1 );
		}
	}

	if( bytes <= 0 )
	{
		error( "bytes: %d <= 0", bytes );
		exit( 0 );
	}
	return bytes; 
}

static void disconnect()
{
	jobControlKillJobs();
}

static time_t getTime()
{
	static time_t start;
	static int firstTime = 1;
	
	if( firstTime )
	{
		firstTime = 0;
		time( &start );
		return difftime( start, start );
	}
	else
	{
		time_t now;
		time( &now );
		return difftime( now, start );
	}
}

static void statusStart( char* bps )
{
	statusDialogInfoRec info;
	int  ptySlaveDev;
	if( openPty( &ptyMasterDev, &ptySlaveDev ) < 0 )
	{
		error( "openPty: %m" );
		exit( 1 );
	}
	jobControlNewJob( localProgName, ptySlaveDev );
	info.fd1 = modemDevice;
	info.fd2 = ptyMasterDev;
	info.readCallback = doRead;
	info.disconnectCallback = disconnect;
	info.timeCallback = getTime;
	info.connectSpeed = bps; 

	statusDialogInit( &info );
	statusDialogPopup();
}

static void xmModemExit()
{
	jobControlKillJobs();

	close( modemDevice );
	close( ptyMasterDev );
	fprintf(stderr, "Bye" );
}

void main( int argc, char* argv[] )
{
	Widget toplevel;
	atexit( xmModemExit );
	
	toplevel = interfaceCreate( &argc, argv, "XMmdm" );

	setErrorFunc( xmmdm_error );
	setModemReadFunc( protReceive );
	setModemWriteFunc( protSend );
	signal( SIGPIPE, signalHandler );
	
	initStart( argc, argv );

	modemDevice = modemd_connect();

	modemRead( modemDevice, &commands, sizeof( commands ) );
#ifdef DEBUG
	test();
#endif
	
	mdm_set_commands( &commands );

	dialDialogCreate( toplevel );
	dialingDialogCreate( toplevel );
	statusDialogCreate( toplevel );
	
	interface = 1;
	
	jobControlInit();
	
	getNumberStart( phonenumber );
	interfaceRun();
}
		
/* EOF main.c */
